/***********************************************************************
*
* mimedefang.c
*
* C interface to the attachment-filter program for stripping or altering
* MIME attachments in incoming Sendmail connections.
*
* Copyright (C) 2000 Roaring Penguin Software Inc.
* http://www.roaringpenguin.com
*
* This program may be distributed under the terms of the GNU General
* Public License, Version 2, or (at your option) any later version.
*
* This program was derived from the sample mail filter included in
* libmilter/README in the Sendmail 8.10 distribution.
***********************************************************************/

static char const RCSID[] =
"$Id: mimedefang.c,v 1.4 2000/08/28 19:41:56 dfs Exp $";

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include "libmilter/mfapi.h"
#include "sendmail/sysexits.h"

/* Keep track of private data -- file name and fp for writing e-mail body */

struct privdata {
    char *dir;
    FILE *fp;
    FILE *headerFP;
    FILE *recipFP;
};

static sfsistat cleanup(SMFICTX *ctx);

int DirCount = 0;
#define DATA ((struct privdata *) smfi_getpriv(ctx))

/* How many tries to get a unique directory name before giving up? */
#define MAXTRIES 8192

/* Size of buffer for temporary string manipulation */
#define SMALLBUF 256

/* Size of chunk when replacing body */
#define CHUNK 4096

/**********************************************************************
*%FUNCTION: envfrom
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
* from -- list of arguments to "MAIL FROM:" SMTP command.
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Allocates a private data structure for tracking this message, and
* opens a temporary file for saving message body.
***********************************************************************/
static sfsistat
envfrom(SMFICTX *ctx, char **from)
{
    struct privdata *data;
    int fd;
    int tries;
    int success;
    char buffer[SMALLBUF];
    char buf2[SMALLBUF];
    FILE *fp;

    /* We need a nice clean directory all our own.  It may seem that
     * using DirCount in this way is dangerous because of thread-safety,
     * but it doesn't really matter.  We just need something that
     * increments, and if we miss a number or two, who cares?  If two
     * threads get the same number, the pthread_self() will disambiguate
     * them. */

    success = 0;
    for (tries = 0; tries < MAXTRIES; tries++) {
	++DirCount;
	snprintf(buffer, SMALLBUF, "/tmp/mdefang-%ld-%d", 
		 (long) pthread_self(), DirCount);
	if (!mkdir(buffer, 0700)) {
	    success = 1;
	    break;
	}
	if (errno != EEXIST) {
	    break;
	}
    }

    /* Could not create temp. directory */
    if (!success) {
	return SMFIS_TEMPFAIL;
    }
    
    data = malloc(sizeof *data);
    if (!data) {
	return SMFIS_TEMPFAIL;
    }

    data->dir = NULL;
    data->fp = NULL;
    data->headerFP = NULL;
    data->recipFP = NULL;

    data->dir = strdup(buffer);

    if (!data->dir) {
	free(data);
	return SMFIS_TEMPFAIL;
    }

    /* Synthesize filename for output */
    snprintf(buf2, SMALLBUF, "%s/INPUTMSG", buffer);
    if ((fd = open(buf2, O_CREAT | O_EXCL | O_APPEND | O_RDWR, 0600)) < 0 ||
	(data->fp = fdopen(fd, "w+")) == NULL) {
	free(data->dir);
	free(data);
	return SMFIS_TEMPFAIL;
    }

    snprintf(buf2, SMALLBUF, "%s/HEADERS", buffer);
    if ((fd = open(buf2, O_CREAT | O_EXCL | O_APPEND | O_RDWR, 0600)) < 0 ||
	(data->headerFP = fdopen(fd, "w+")) == NULL) {
	fclose(data->fp);
	free(data->dir);
	free(data);
	return SMFIS_TEMPFAIL;
    }

    snprintf(buf2, SMALLBUF, "%s/RECIPIENTS", buffer);
    if ((fd = open(buf2, O_CREAT | O_EXCL | O_APPEND | O_RDWR, 0600)) < 0 ||
	(data->recipFP = fdopen(fd, "w+")) == NULL) {
	fclose(data->headerFP);
	fclose(data->fp);
	free(data->dir);
	free(data);
	return SMFIS_TEMPFAIL;
    }

    /* Write the sender */
    snprintf(buf2, SMALLBUF, "%s/SENDER", data->dir);
    if ((fd = open(buf2, O_CREAT | O_EXCL | O_APPEND | O_RDWR, 0600)) < 0 ||
	(fp = fdopen(fd, "w+")) == NULL) {
	fclose(data->recipFP);
	fclose(data->headerFP);
	fclose(data->fp);
	free(data->dir);
	free(data);
	return SMFIS_TEMPFAIL;
    }
    fprintf(fp, "%s\n", from[0]);
    fclose(fp);

    /* Save private data */
    smfi_setpriv(ctx, data);
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: rcptto
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
* to -- list of arguments to each RCPT_TO
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Saves recipient data
***********************************************************************/
static sfsistat
rcptto(SMFICTX *ctx, char **to)
{
    FILE *fp;
    fprintf(DATA->recipFP, "%s\n", to[0]);
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: header
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
* headerf -- Header field name
* headerv -- Header value
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Writes the header to the temporary file
***********************************************************************/
static sfsistat
header(SMFICTX *ctx, char *headerf, char *headerv)
{
    fprintf(DATA->fp, "%s: %s\r\n", headerf, headerv);
    fprintf(DATA->headerFP, "%s: %s\r\n", headerf, headerv);
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: eoh
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Writes a blank line to indicate the end of headers.
***********************************************************************/
static sfsistat
eoh(SMFICTX *ctx)
{
    fprintf(DATA->fp, "\r\n");
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: body
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
* text -- a chunk of text from the mail body
* len -- length of chunk
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Writes a chunk of the body to the temporary file
***********************************************************************/
static sfsistat
body(SMFICTX *ctx, u_char *text, size_t len)
{
    if (fwrite(text, len, 1, DATA->fp) <= 0) {
	cleanup(ctx);
	return SMFIS_TEMPFAIL;
    }
    return SMFIS_CONTINUE;
}

/**********************************************************************
*%FUNCTION: eom
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* This is where all the action happens.  Called at end of message, it
* forks the Perl scanner which may or may not ask for the body to be
* replaced.
***********************************************************************/
static sfsistat
eom(SMFICTX *ctx)
{
    pid_t n;
    int status;
    char buffer[SMALLBUF];
    FILE *fp;
    struct privdata *data = DATA;

    /* Close output file */
    fclose(data->fp);
    fclose(data->headerFP);
    fclose(data->recipFP);
    data->fp = NULL;
    data->headerFP = NULL;
    data->recipFP = NULL;

    /* Create a sanity-check file */
    snprintf(buffer, SMALLBUF, "%s/UNFINISHED", data->dir);
    if ((fp = fopen(buffer, "w")) != NULL) {
	fclose(fp);
    }

    /* Run the Perl filter */
    n = fork();
    if (n < 0) {
	cleanup(ctx);
	return SMFIS_TEMPFAIL;
    } else if (n == 0) {
	/* In the child */
	execl(PERL_PATH, "perl", MIMEDEFANG_PL, data->dir, NULL);
	_exit(1);
    } else {
	/* In the parent */
	if (waitpid(n, &status, 0) < 0) {
	    cleanup(ctx);
	    return SMFIS_TEMPFAIL;
	}
    }

    /* If BOUNCE file is there, bounce the message */
    snprintf(buffer, SMALLBUF, "%s/BOUNCE", data->dir);
    if (!access(buffer, F_OK)) {
	int fd, n;
	fd = open(buffer, O_RDONLY);
	if (fd < 0) {
	    cleanup(ctx);
	    return SMFIS_TEMPFAIL;
	}
	if ((n=read(fd, buffer, SMALLBUF-1)) < 0) {
	    close(fd);
	    cleanup(ctx);
	    return SMFIS_TEMPFAIL;
	}
	close(fd);
	if (!n) ++n;
	buffer[n] = 0;
	smfi_setreply(ctx, "554", "5.7.1", buffer);
	cleanup(ctx);
	return SMFIS_REJECT;
    }

    /* If body is changed, replace it */
    snprintf(buffer, SMALLBUF, "%s/CHANGED", data->dir);
    if (!access(buffer, F_OK)) {
	int fd;
	int j;
	char chunk[CHUNK];
	snprintf(buffer, SMALLBUF, "%s/NEWBODY", data->dir);
	fd = open(buffer, O_RDONLY);
	if (fd < 0) {
	    cleanup(ctx);
	    return SMFIS_TEMPFAIL;
	}
	while ((j=read(fd, chunk, CHUNK)) > 0) {
	    smfi_replacebody(ctx, chunk, j);
	}
	close(fd);
    }

    /* If UNFINISHED file is still there, ooops... */
    snprintf(buffer, SMALLBUF, "%s/UNFINISHED", data->dir);
    if (!access(buffer, F_OK)) {
	smfi_addheader(ctx, "X-MIMEDefang-Warning", "Perl process did not finish");
    }
    smfi_addheader(ctx, "X-Scanned-By", "MIMEDefang " VERSION " (http://www.roaringpenguin.com/mimedefang/)");
    return cleanup(ctx);
}

/**********************************************************************
*%FUNCTION: mfclose
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
*%RETURNS:
* SMFIS_ACCEPT
*%DESCRIPTION:
* Called when connection is closed.
***********************************************************************/
static sfsistat
mfclose(SMFICTX *ctx)
{
    return SMFIS_ACCEPT;
}

/**********************************************************************
*%FUNCTION: mfabort
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Called if connection is aborted.  Just cleans up.
***********************************************************************/
static sfsistat
mfabort(SMFICTX *ctx)
{
    return cleanup(ctx);
}

/**********************************************************************
*%FUNCTION: cleanup
*%ARGUMENTS:
* ctx -- Sendmail filter mail context
*%RETURNS:
* SMFIS_TEMPFAIL or SMFIS_CONTINUE
*%DESCRIPTION:
* Cleans up temporary files.
***********************************************************************/
static sfsistat
cleanup(SMFICTX *ctx)
{
    sfsistat r = SMFIS_CONTINUE;
    struct privdata *data = DATA;
    pid_t n;

    if (!data) return r;

    if (data->fp && (fclose(data->fp) == EOF)) {
	r = SMFIS_TEMPFAIL;
    }

    if (data->headerFP && (fclose(data->headerFP) == EOF)) {
	r = SMFIS_TEMPFAIL;
    }

    if (data->recipFP && (fclose(data->recipFP) == EOF)) {
	r = SMFIS_TEMPFAIL;
    }

    /* Remove temporary files */
    n = fork();
    if (n < 0) {
	r = SMFIS_TEMPFAIL;
    } else if (n == 0) {
	/* In the child */
	execl(RM, "rm", "-rf", data->dir, NULL);
	_exit(1);
    } else {
	/* In the parent */
	if (waitpid(n, NULL, 0) < 0) {
	    r = SMFIS_TEMPFAIL;
	}
    }
    free(data->dir);
    free(data);
    smfi_setpriv(ctx, NULL);
    return r;
}

static struct smfiDesc filterDescriptor =
{
    "MIMEDefang-" VERSION,      /* Filter name */
    SMFI_VERSION,		/* Version code */

#if SMFI_VERSION == 1
    SMFIF_MODHDRS|SMFIF_MODBODY, /* We add a header and may alter body */
#else
    SMFIF_ADDHDRS|SMFIF_MODBODY, /* We add a header and may alter body */
#endif

    NULL,			/* connection */
    NULL,			/* HELO */
    envfrom,			/* MAIL FROM: */
    rcptto,			/* RCPT TO: */
    header,			/* Called for each header */
    eoh,			/* Called at end of headers */
    body,			/* Called for each body chunk */
    eom,			/* Called at end of message */
    mfabort,			/* Called on abort */
    mfclose			/* Called on connection close */
};

/**********************************************************************
* %FUNCTION: main
* %ARGUMENTS:
*  argc, argv -- the usual suspects
* %RETURNS:
*  Whatever smfi_main returns
* %DESCRIPTION:
*  Main program
***********************************************************************/
int
main(int argc, char **argv)
{
    char c;

    /* Paranoia time */
    umask(077);
    /* Process command line options */
    while ((c = getopt(argc, argv, "p:")) != -1) {
	switch (c) {
	case 'p':
	    if (optarg == NULL || *optarg == '\0') {
		(void) fprintf(stderr, "Illegal conn: %s\n",
			       optarg);
		exit(EX_USAGE);
	    }
	    if (smfi_setconn(optarg) != MI_SUCCESS) {
		fprintf(stderr, "Could not open connection %s: %s",
			optarg, strerror(errno));
		exit(1);
	    }
	    break;
	}
    }
    if (smfi_register(filterDescriptor) == MI_FAILURE) {
	fprintf(stderr, "smfi_register failed\n");
	exit(EX_UNAVAILABLE);
    }
    return smfi_main();
}
